//
// hc165.ino
//

// This sketch demonstrates how to control cascade-connected two 74HC165 ICs. The following conditions are assumed:
//
// (1) The Arduino's pin 5 is connected to the QH pin of the last stage 74HC165 IC.
// (2) The Arduino's pin 6 is connected to the CLK pins of all 74HC165 ICs.
// (3) The Arduino's pin 7 is connected to the SH//LD pins of all 74HC165 ICs.
//
//  You can use 16 input pins with the expense of only three I/O pins of the Arduino. If you use more 74HC165 ICs, 
//  you can use more input pins with no extra Arduino's I/O pins.

// Pin layout of 74HC165 in DIP package.
//
//              ------------              
//         ----|    |  |    |----         
// SH//LD |  1 |     --     | 16 | VCC    
//         ----|            |----         
//         ----|            |----         
//    CLK |  2 |            | 15 | CLK INH
//         ----|            |----         
//         ----|            |----         
//      E |  3 |            | 14 | D      
//         ----|            |----         
//         ----|            |----         
//      F |  4 |            | 13 | C      
//         ----|            |----         
//         ----|            |----         
//      G |  5 |            | 12 | B      
//         ----|            |----         
//         ----|            |----         
//      H |  6 |            | 11 | A      
//         ----|            |----         
//         ----|            |----         
//    /QH |  7 |            | 10 | SER    
//         ----|            |----         
//         ----|            |----         
//    GND |  8 |            |  9 | QH     
//         ----|            |----         
//              ------------              


// Descriptions of 74HC165 pins
//
// No.  Name    Description
//  1   SH//LD  Selection signal for the shift mode and load mode. if HIGH, shift mode is selected. Connect this pin to the Arduino's I/O pin.
//  2   CLK     Shift clock. Connect this pin to the Arduino's I/O pin.
//  3   E       Data input pin. Bit3 in LSBFIRST mode.  Bit4 in MSBFIRST mode.
//  4   F       Data input pin. Bit2 in LSBFIRST mode.  Bit5 in MSBFIRST mode.
//  5   G       Data input pin. Bit1 in LSBFIRST mode.  Bit6 in MSBFIRST mode.
//  6   H       Data input pin. Bit0 (LSB) in LSBFIRST mode.  Bit7 (MSB) in MSBFIRST mode.
//  7   /QH     Inverted Serial output pin. Usually this pin is not used. Leave this pin open.
//  8   GND     Ground pin. Connect this pin to the GND pin of the Arduino.
//  9   QH      Serial outpu pin. If this is not the last stage 74HC165, connect this QH pin to the SER pin of the next stage 74HC165. If this is
//              the last stage 74HC165, connect this QH pin to the Arduino's I/O pin.
// 10   SER     Serial input pin. If this is the first stage 74HC165, connect this SER pin to either VCC or GND. Don't leave this pin open. If 
//              this is not the first stage 74HC165, connect this SER pin to the QH pin of the previous stage 74HC165.
// 11   A       Data input pin. Bit7 (MSB) in LSBFIRST mode.  Bit0 (LSB) in MSBFIRST mode.
// 12   B       Data input pin. Bit6 in LSBFIRST mode.  Bit1 in MSBFIRST mode.
// 13   C       Data input pin. Bit5 in LSBFIRST mode.  Bit2 in MSBFIRST mode.
// 14   D       Data input pin. Bit4 in LSBFIRST mode.  Bit3 in MSBFIRST mode.
// 15   CLK INH A signal to inhivit shifit clock. Usually this pin is not used. Connect this pin to GND.
// 16   VCC     Positive power supply pin. Connect this pin to the 5V pin (if 5V Arduino) or the 3.3V pin (if 3.3V Arduino) of the Arduino.

#include <fastio.h>

hc165<5,6,7,2> my165; // Declares an object that controls cascade-connected two 74HC165s. The parameters are the pin for QH, the pin for CLK, the
                      // pin for SH//LD and the number of the 74HC165 ICs respectively.

hc165pin<decltype(my165), my165,  5> pin5 ; // Gives the name, "pin5" to the bit 5 of 74HC165 ICs.
hc165pin<decltype(my165), my165, 14> pin14; // Gives the name, "pin14" to the bit 14 of 74HC165 ICs.

void setup()
{
  my165.init(); // Initiallizes the 74HC165 driver. Set the pins for CLK and SH//LD pins OUTPUTs. Set the pin for QH INPUT.
  Serial.begin(9600); // Opens serial port at 9600bps. 
} // setup

void loop()
{
  uint16_t data;

  data=my165.shiftIn(); // reads 16 bits from two 74HC165s.
  Serial.println(data,HEX);

  delay(1000);

  data=my165; // reads 16 bits from two 74HC165s.
  Serial.println(data,HEX);

  delay(1000);

  data=my165.digitalRead<3>(); // reads bit 3.
  Serial.print("Bit 3 is ");
  Serial.print(data);
  Serial.println('.');

  delay(1000);

  data=pin5.digitalRead(); // reads bit 5.
  Serial.print("Bit 5 is ");
  Serial.print(data);
  Serial.println('.');

  delay(1000);

  data=pin14; // reads bit 14.
  Serial.print("Bit 14 is ");
  Serial.print(data);
  Serial.println('.');

  delay(1000);

  // Reads bit 3, 5 and 14 at the same time.
  my165.fetch(); // Reads all bits to the internal buffer
  
  data=my165.readFromBuffer<3>(); // Reads bit 3 from the internal buffer. No transfer occurs now.
  Serial.print("Bit 3 is ");
  Serial.print(data);
  Serial.println('.');

  data=pin5.readFromBuffer(); // Reads bit 5 from the internal buffer. No transfer occurs now.
  Serial.print("Bit 5 is ");
  Serial.print(data);
  Serial.println('.');

  data=pin14.readFromBuffer(); // Reads bit 14 from the internal buffer. No transfer occurs now.
  Serial.print("Bit 14 is ");
  Serial.print(data);
  Serial.println('.');

  Serial.println();
  delay(3000);
} // loop
