//
// hc595.ino
//

// This sketch demonstrates how to control cascade-connected two 74HC595 ICs. The following conditions are assumed:
//
// (1)The Arduino's pin 2 is connected to the SER pin of the first stage 74HC595 IC.
// (2)The Arduino's pin 3 is connected to the SRCLK pins of all 74HC595 ICs.
// (3)The Arduino's pin 4 is connected to the RCL pins of all 74HC595 ICs.
//
// You can use 16 output pins with the expense of only three I/O pins of the Arduino. If you use more 74HC165 ICs,
// you can use more output pins with no extra Arduino's I/O pins.

// Pin layout of 74HC595 in DIP package.
//
//           ------------             
//      ----|    |  |    |----        
//  QB |  1 |     --     | 16 | VCC   
//      ----|            |----        
//      ----|            |----        
//  QC |  2 |            | 15 | QA    
//      ----|            |----        
//      ----|            |----        
//  QD |  3 |            | 14 | SER   
//      ----|            |----        
//      ----|            |----        
//  QE |  4 |            | 13 | /OE   
//      ----|            |----        
//      ----|            |----        
//  QF |  5 |            | 12 | RCLK  
//      ----|            |----        
//      ----|            |----        
//  QG |  6 |            | 11 | SRCLK 
//      ----|            |----        
//      ----|            |----        
//  QH |  7 |            | 10 | /SRCLR
//      ----|            |----        
//      ----|            |----        
// GND |  8 |            |  9 | QH'   
//      ----|            |----        
//           ------------


// Descriptions of 74HC595 pins
//
// No.  Name   Description
//  1   QB     Data output pin.  Bit6 in LSBFIRST mode. Bit1 in MSBFIRST mode.
//  2   QC     Data output pin.  Bit5 in LSBFIRST mode. Bit2 in MSBFIRST mode.
//  3   QD     Data output pin.  Bit4 in LSBFIRST mode. Bit3 in MSBFIRST mode.
//  4   QE     Data output pin.  Bit3 in LSBFIRST mode. Bit4 in MSBFIRST mode.
//  5   QF     Data output pin.  Bit2 in LSBFIRST mode. Bit5 in MSBFIRST mode.
//  6   QG     Data output pin.  Bit1 in LSBFIRST mode. Bit6 in MSBFIRST mode.
//  7   QH     Data output pin.  Bit0 (LSB) in LSBFIRST mode. Bit7 (MSB) in MSBFIRST mode.
//  8   GND    Ground pin. Connect this pin to the GND pin of the Arduino.
//  9   QH'    Unregistered output of QH. Connect this pin to the SER pin of the next stage 74HC595. If this is the last stage 74HC595, leave this QH' pin open.
// 10   /SRCLR Negative input to clear the shift register. Connect this pin to VCC.
// 11   SRCLK  Shift clock input. Connect this pin to the Arduino.
// 12   RCLK   Latch clock for the strage regisetr. Connect this pin to the Arduino's I/O pin.
// 13   /OE    Negative input of output enable. Connect this pin to VCC.
// 14   SER    Serial input pin of the shift register. If this is the first stage 74HC595, connect this SER pin to the Arduino's I/O pin. If this is not the first 
//             stage 74HC595, connect this SER ppin to the QH' pin of the previous stage 74HC595.
// 15   QA     Data output pin.  Bit7 (MSB) in LSBFIRST mode. Bit0 (LSB) in MSBFIRST mode.
// 16   VCC    Positive power supply pin. Connect this pin to the 5V pin (if 5V Arduino) or the 3.3V pin (if 3.3V Arduino) of the Arduino.

#include <fastio.h>

hc595<2,3,4,2> my595; // Declares an object that controls cascade-connected two 74HC595s. The parameters are the pin for SER, the pin for SRCLK , the pin for 
                      // RCLK and the number of the 74HC595 ICs respectively.

hc595pin<decltype(my595), my595, 1> pin1; // Gives the name, "pin1" to the bit 1 of 74HC595 ICs.
hc595pin<decltype(my595), my595, 2> pin2; // Gives the name, "pin2" to the bit 2 of 74HC595 ICs.

void setup()
{
  my595.init(); // Initializes the 74HC595s and their driver.  Sets the pins for SER, SRCLK and RCL signals OUTPUT, clears data buffer, and send 16-bits 0s.
} // setup

void loop()
{
  my595.shiftOut(0x1234); // The easiest way to use this library as a 74HC595 driver. Sends 16-bit data, 0x1234.
  my595=0x1234; // Abbreviated expresson for the previous command.

  delay(1000);

  my595.digitalWrite<0>(HIGH); // Sets bit 0 HIGH immediately.

  delay(1000);
  
  pin1.digitalWrite(HIGH); // Sets bit 1 HIGH immediately.
  
  delay(1000);
  
  pin2=LOW; // Sets bit 2 LOW immediately. (abbreviated expression)

  my595.writeToBuffer<0>(LOW); // Reserves future write that sets bit 0 LOW. (Nothing happens immediately.)

  delay(1000);
  
  pin1.writeToBuffer(LOW); // Reserves future write that sets bit 1 LOW. (Nothing happens immediately.)

  delay(1000);
  
  pin2=HIGH; // Changes bit 0, bit 1 and bit 2 at the same time.

  delay(1000);

  my595=0; // Set all bits 0s.

  delay(2000);
} // loop
